from supervisely.app.widgets.widget import Widget
from supervisely.app.widgets.radio_table.radio_table import RadioTable
from supervisely.app.widgets.notification_box.notification_box import NotificationBox
from supervisely.app.widgets.done_label.done_label import DoneLabel
from supervisely.app.widgets.sly_tqdm.sly_tqdm import SlyTqdm, Progress
from supervisely.app.widgets.grid_gallery.grid_gallery import GridGallery
from supervisely.app.widgets.classic_table.classic_table import ClassicTable
from supervisely.app.widgets.confusion_matrix.confusion_matrix import ConfusionMatrix
from supervisely.app.widgets.project_selector.project_selector import ProjectSelector
from supervisely.app.widgets.element_button.element_button import ElementButton
from supervisely.app.widgets.button.button import Button
from supervisely.app.widgets.project_thumbnail.project_thumbnail import ProjectThumbnail
from supervisely.app.widgets.apexchart.apexchart import Apexchart
from supervisely.app.widgets.line_chart.line_chart import LineChart
from supervisely.app.widgets.heatmap_chart.heatmap_chart import HeatmapChart
from supervisely.app.widgets.table.table import Table
from supervisely.app.widgets.labeled_image.labeled_image import LabeledImage
from supervisely.app.widgets.text.text import Text

# from supervisely.app.widgets.restart_step.restart_step import RestartStep
